package cz.muni.fi.pv168.invoicemanager.backend;

import cz.muni.fi.pv168.common.DBUtils;
import cz.muni.fi.pv168.common.IllegalEntityException;
import cz.muni.fi.pv168.common.ValidationException;
import java.sql.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sql.DataSource;

/**
 * Implementation of InvoiceManager service.
 * 
 * @author Petr Adámek
 */
public class InvoiceManagerImpl implements InvoiceManager {

    private static final Logger logger = Logger.getLogger(
            InvoiceManagerImpl.class.getName());
    
    private DataSource dataSource;

    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
    }
    
    private void checkDataSource() {
        if (dataSource == null) {
            throw new IllegalStateException("DataSource is not set");
        }
    }    
    
    @Override
    public List<Invoice> findAllInvoices() throws ServiceFailureException {
        checkDataSource();
        Connection conn = null;
        PreparedStatement st = null;
        try {
            conn = dataSource.getConnection();
            st = conn.prepareStatement(
                    "SELECT id, number, customer, issued FROM Invoice");
            ResultSet rs = st.executeQuery();
            List<Invoice> result = new ArrayList<Invoice>();
            while (rs.next()) {
                Invoice invoice = rowToInvoice(rs);
                loadInvoiceRows(conn, invoice);
                result.add(invoice);
            }
            return result;
        } catch (SQLException ex) {
            String msg = "Error when getting all invoices from DB";
            logger.log(Level.SEVERE, msg, ex);
            throw new ServiceFailureException(msg, ex);
        } finally {
            DBUtils.closeQuietly(conn, st);
        }          
    }

    @Override
    public void createInvoice(Invoice invoice) throws ServiceFailureException {
        checkDataSource();
        validate(invoice);
        if (invoice.getId() != null) {
            throw new IllegalEntityException("invoice id is already set");
        }        
        Connection conn = null;
        PreparedStatement st = null;
        try {
            conn = dataSource.getConnection();
            // Temporary turn autocommit mode off. It is turned back on in 
            // method DBUtils.closeQuietly(...) 
            conn.setAutoCommit(false); 
            st = conn.prepareStatement(
                    "INSERT INTO Invoice (number,customer,issued) VALUES (?,?,?)",
                    Statement.RETURN_GENERATED_KEYS);
            st.setString(1, invoice.getNumber());
            st.setString(2, invoice.getCustomer());
            st.setTimestamp(3, dateToTimestamp(invoice.getIssued()));

            int count = st.executeUpdate();
            DBUtils.checkUpdatesCount(count, invoice, true);

            Long id = DBUtils.getId(st.getGeneratedKeys());
            invoice.setId(id);
            
            storeInvoiceRows(conn, invoice);
            
            conn.commit();
            
        } catch (SQLException ex) {
            String msg = "Error when inserting invoice into db";
            logger.log(Level.SEVERE, msg, ex);
            throw new ServiceFailureException(msg, ex);
        } finally {
            DBUtils.doRollbackQuietly(conn);
            DBUtils.closeQuietly(conn, st);
        }
    }

    private Timestamp dateToTimestamp(Date date) {
        if (date == null) {
            return null;
        } else if (date instanceof Timestamp) {
            return (Timestamp) date;
        } else {
            return new Timestamp(date.getTime());
        }
    }

    @Override
    public Invoice getInvoice(Long id) throws ServiceFailureException {
        
        checkDataSource();
        
        if (id == null) {
            throw new IllegalArgumentException("id is null");
        }
        
        Connection conn = null;
        PreparedStatement st = null;
        try {
            conn = dataSource.getConnection();
            st = conn.prepareStatement(
                    "SELECT id, number, customer, issued FROM Invoice WHERE id = ?");
            st.setLong(1, id);
            ResultSet rs = st.executeQuery();
            if (rs.next()) {
                Invoice result = rowToInvoice(rs);                
                if (rs.next()) {
                    throw new ServiceFailureException(
                            "Internal integrity error: more invoices with the same id found!");
                }
                loadInvoiceRows(conn, result);
                return result;
            } else {
                return null;
            }
        } catch (SQLException ex) {
            String msg = "Error when getting invoice with id = " + id + " from DB";
            logger.log(Level.SEVERE, msg, ex);
            throw new ServiceFailureException(msg, ex);
        } finally {
            DBUtils.closeQuietly(conn, st);
        }
    }

    @Override
    public void updateInvoice(Invoice invoice) throws ServiceFailureException {
        checkDataSource();
        validate(invoice);
        if (invoice.getId() == null) {
            throw new IllegalEntityException("invoice id is null");
        }        
        Connection conn = null;
        PreparedStatement st = null;
        try {
            conn = dataSource.getConnection();
            // Temporary turn autocommit mode off. It is turned back on in 
            // method DBUtils.closeQuietly(...) 
            conn.setAutoCommit(false);
            st = conn.prepareStatement(
                    "UPDATE Invoice SET number = ?, customer = ?, issued = ? WHERE id = ?");
            st.setString(1, invoice.getNumber());
            st.setString(2, invoice.getCustomer());
            st.setTimestamp(3, dateToTimestamp(invoice.getIssued()));
            st.setLong(4, invoice.getId());

            int count = st.executeUpdate();
            DBUtils.checkUpdatesCount(count, invoice, false);

            updateInvoiceRows(conn, invoice);
            
            conn.commit();
            
        } catch (SQLException ex) {
            String msg = "Error when updating invoice in the db";
            logger.log(Level.SEVERE, msg, ex);
            throw new ServiceFailureException(msg, ex);
        } finally {
            DBUtils.doRollbackQuietly(conn);
            DBUtils.closeQuietly(conn, st);
        }        
    }

    @Override
    public void deleteInvoice(Invoice invoice) throws ServiceFailureException {
        checkDataSource();
        if (invoice == null) {
            throw new IllegalArgumentException("invoice is null");
        }        
        if (invoice.getId() == null) {
            throw new IllegalEntityException("invoice id is null");
        }        
        Connection conn = null;
        PreparedStatement st = null;
        try {
            conn = dataSource.getConnection();
            // Temporary turn autocommit mode off. It is turned back on in 
            // method DBUtils.closeQuietly(...) 
            conn.setAutoCommit(false);
            st = conn.prepareStatement(
                    "DELETE FROM InvoiceRow WHERE invoiceId = ?");
            st.setLong(1, invoice.getId());
            st.executeUpdate();
            
            st = conn.prepareStatement(
                    "DELETE FROM Invoice WHERE id = ?");
            st.setLong(1, invoice.getId());

            int count = st.executeUpdate();
            DBUtils.checkUpdatesCount(count, invoice, false);

            conn.commit();

        } catch (SQLException ex) {
            String msg = "Error when deleting invoice from the db";
            logger.log(Level.SEVERE, msg, ex);
            throw new ServiceFailureException(msg, ex);
        } finally {
            DBUtils.doRollbackQuietly(conn);
            DBUtils.closeQuietly(conn, st);
        }
    }

    private static void loadInvoiceRows(Connection conn, Invoice invoice) throws SQLException {
        if (!invoice.getRows().isEmpty()) {
            throw new IllegalArgumentException("invoice alrady have some rows!");
        }
        PreparedStatement st = null;
        try {
            st = conn.prepareStatement(
                    "SELECT id, invoiceId, description, amount, unitPrice FROM InvoiceRow WHERE invoiceId = ?");
            st.setLong(1, invoice.getId());
            ResultSet rs = st.executeQuery();
            while (rs.next()) {
                invoice.getRows().add(rowToInvoiceRow(rs,invoice));
            }
        } finally {
            DBUtils.closeQuietly(null, st);
        }
    }

    private static PreparedStatement prepareInsertInvoiceRowStatement(Connection conn) throws SQLException {
        return conn.prepareStatement(
                "INSERT INTO InvoiceRow (invoiceId, description, amount, unitPrice) VALUES (?,?,?,?)",
                Statement.RETURN_GENERATED_KEYS);
    }

    private static void callInsertInvoiceRowStatement(PreparedStatement insertSt, InvoiceRow row) throws SQLException {
        if (row.getId() != null) {
            throw new IllegalEntityException("InvoiceRow " + row + " have already asigned some id");
        }
        insertSt.setLong(1,row.getInvoice().getId());
        insertSt.setString(2,row.getDescription());
        insertSt.setBigDecimal(3,row.getAmount());
        insertSt.setBigDecimal(4,row.getUnitPrice());
        int updatesCount = insertSt.executeUpdate();
        DBUtils.checkUpdatesCount(updatesCount, row, true);
        row.setId(DBUtils.getId(insertSt.getGeneratedKeys()));
    }

    private static PreparedStatement prepareUpdateInvoiceRowStatement(Connection conn) throws SQLException {
        return conn.prepareStatement(
                "UPDATE InvoiceRow SET invoiceId = ?, description = ?, amount = ?, unitPrice = ? WHERE id = ?");
    }

    private static void callUpdateInvoiceRowStatement(PreparedStatement updateSt, InvoiceRow row) throws SQLException {
        if (row.getId() == null) {
            throw new IllegalEntityException("InvoiceRow have no id");
        }
        updateSt.setLong(1,row.getInvoice().getId());
        updateSt.setString(2,row.getDescription());
        updateSt.setBigDecimal(3,row.getAmount());
        updateSt.setBigDecimal(4,row.getUnitPrice());
        updateSt.setLong(5,row.getId());
        int updatesCount = updateSt.executeUpdate();
        DBUtils.checkUpdatesCount(updatesCount, row, false);
    }

    private static String questionMarkList(int count) {
        StringBuilder result = new StringBuilder(count * 2);
        for (int i = 0; i < count; i++) {
            if (i > 0) {
                result.append(',');
            }
            result.append('?');
        }
        return result.toString();
    }

    private static void deleteRemovedInvoiceRows(Connection conn, Invoice invoice) throws SQLException {
        PreparedStatement deleteSt = null;
        try {
            int rowsCount = invoice.getRows().size();
            String sql = "DELETE FROM InvoiceRow WHERE invoiceId = ?";
            if (rowsCount > 0) {
                sql += " AND id NOT IN (" + questionMarkList(rowsCount) + ")";
            }                
            deleteSt = conn.prepareStatement(sql);
            int i = 1;
            deleteSt.setLong(i++, invoice.getId());            
            for (InvoiceRow row : invoice.getRows()) {
                deleteSt.setLong(i++, row.getId());
            }
            deleteSt.executeUpdate();
        } finally {
            DBUtils.closeQuietly(null, deleteSt);
        }
        
    }
    
    private static void checkInvoiceFieldOnInvoiceRow(Invoice invoice, InvoiceRow row) throws IllegalArgumentException {
        if (row.getInvoice() == null) {
            throw new IllegalEntityException("InvoiceRow " + row + 
                    " has no invoice field set");
        }
        if (invoice != row.getInvoice()) {
            throw new IllegalEntityException("Invoice " + invoice + 
                    " contains row " + row + " that belongs to another invoice "
                    + row.getInvoice());
        }
    }

    private static void storeInvoiceRows(Connection conn, Invoice invoice) throws SQLException {
        PreparedStatement insertSt = null;
        try {
            insertSt = prepareInsertInvoiceRowStatement(conn);
            for (InvoiceRow row : invoice.getRows()) {
                checkInvoiceFieldOnInvoiceRow(invoice, row);
                callInsertInvoiceRowStatement(insertSt, row);
            }
        } finally {
            DBUtils.closeQuietly(null, insertSt);
        }
    }
    
    private static void updateInvoiceRows(Connection conn, Invoice invoice) throws SQLException {
        PreparedStatement insertSt = null, updateSt = null, deleteSt = null;
        try {
            insertSt = prepareInsertInvoiceRowStatement(conn);
            updateSt = prepareUpdateInvoiceRowStatement(conn);
            for (InvoiceRow row : invoice.getRows()) {
                checkInvoiceFieldOnInvoiceRow(invoice, row);
                if (row.getId() == null) {
                    callInsertInvoiceRowStatement(insertSt, row);
                } else {
                    callUpdateInvoiceRowStatement(updateSt, row);
                }
            }
            deleteRemovedInvoiceRows(conn, invoice);
        } finally {
            DBUtils.closeQuietly(null, insertSt, updateSt, deleteSt);
        }        
    }
    
    private static Invoice rowToInvoice(ResultSet rs) throws SQLException {
        Invoice result = new Invoice();
        result.setId(rs.getLong("id"));
        result.setNumber(rs.getString("number"));
        result.setCustomer(rs.getString("customer"));
        result.setIssued(rs.getTimestamp("issued"));
        return result;
        
    }

    private static InvoiceRow rowToInvoiceRow(ResultSet rs, Invoice invoice) throws SQLException {
        InvoiceRow invoiceRow = new InvoiceRow();
        if (rs.getLong("invoiceId") != invoice.getId()) {
            throw new IllegalArgumentException("This row does not correspond to given invoice");
        }
        invoiceRow.setId(rs.getLong("id"));
        invoiceRow.setDescription(rs.getString("description"));
        invoiceRow.setAmount(rs.getBigDecimal("amount"));
        invoiceRow.setUnitPrice(rs.getBigDecimal("unitPrice"));
        invoiceRow.setInvoice(invoice);
        return invoiceRow;
    }
    
    private static void validate(Invoice invoice) {

        if (invoice == null) {
            throw new IllegalArgumentException("invoice is null");
        }
        if (invoice.getNumber() == null) {
            throw new ValidationException("number is null");
        }
        if (invoice.getCustomer() == null) {
            throw new ValidationException("customer is null");
        }
//        if (invoice.getRows().isEmpty()) {
//            throw new ValidationException("invoice has no rows");
//        }
        for (InvoiceRow row : invoice.getRows()) {
            if (row.getDescription() == null) {
                throw new ValidationException("row description is null");
            }
            if (row.getUnitPrice() == null) {
                throw new ValidationException("row unitPrice is null");
            }
            if (row.getUnitPrice().signum() < 0) {
                throw new ValidationException("row unitPrice is negative number");
            }
            if (row.getAmount() == null) {
                throw new ValidationException("row amount is null");
            }
            if (row.getAmount().signum() < 0) {
                throw new ValidationException("row amount is negativ number");
            }
        }
        
    }

}
