package cz.muni.fi.pv168.invoicemanager.backend;

import java.sql.SQLException;
import java.util.Arrays;
import java.util.List;
import javax.sql.DataSource;
import org.apache.commons.dbcp.BasicDataSource;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import cz.muni.fi.pv168.common.DBUtils;
import cz.muni.fi.pv168.common.IllegalEntityException;
import cz.muni.fi.pv168.common.ValidationException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import static org.junit.Assert.*;

/**
 *
 * @author Petr Adámek
 */
public class InvoiceManagerImplTest {
    
    private InvoiceManagerImpl manager;
    private DataSource ds;

    private static DataSource prepareDataSource() throws SQLException {
        BasicDataSource ds = new BasicDataSource();
        //we will use in memory database
        ds.setUrl("jdbc:derby:memory:invoicemgr-test;create=true");
        return ds;
    }

    @Before
    public void setUp() throws SQLException {
        ds = prepareDataSource();
        DBUtils.executeSqlScript(ds,InvoiceManagerImpl.class.getResource("createTables.sql"));
        manager = new InvoiceManagerImpl();
        manager.setDataSource(ds);
    }

    @After
    public void tearDown() throws SQLException {
        DBUtils.executeSqlScript(ds,InvoiceManagerImpl.class.getResource("dropTables.sql"));
    }

    private static final DateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-mm-dd");
    private Date date(String date) {
        try {
            return DATE_FORMAT.parse(date);
        } catch (ParseException ex) {
            throw new RuntimeException(ex);
        }
    }
    
    @Test
    public void createInvoice() {
        Invoice invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
    
        BigDecimal expectedPrice = 
                BigDecimal.valueOf(13.234 * 24.52).setScale(2,RoundingMode.HALF_UP).add(
                BigDecimal.valueOf(4.234 * 219).setScale(2,RoundingMode.HALF_UP));
        
        assertEquals(expectedPrice,invoice.getTotalPrice());
        
        manager.createInvoice(invoice);

        Long invoiceId = invoice.getId();
        assertNotNull(invoiceId);
        Invoice result = manager.getInvoice(invoiceId);
        assertEquals(invoice, result);
        assertNotSame(invoice, result);
        assertInvoiceDeepEquals(invoice, result);
        
    }
    
    @Test
    public void getInvoice() {
        
        assertNull(manager.getInvoice(1l));
        
        Invoice invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));

        manager.createInvoice(invoice);
        Long invoiceId = invoice.getId();

        Invoice result = manager.getInvoice(invoiceId);
        assertEquals(invoice, result);
        assertInvoiceDeepEquals(invoice, result);
    }

    @Test
    public void findAllInvoices() {

        assertTrue(manager.findAllInvoices().isEmpty());

        Invoice i1 = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        Invoice i2 = newInvoice(
                "00034/2012","První stavební", date("2012-08-12"),
                newInvoiceRow("Cement", BigDecimal.valueOf(25), BigDecimal.valueOf(8925,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(1234, 3), BigDecimal.valueOf(219)));

        manager.createInvoice(i1);
        manager.createInvoice(i2);

        List<Invoice> expected = Arrays.asList(i1,i2);
        List<Invoice> actual = manager.findAllInvoices();
        
        assertInvoiceCollectionDeepEquals(expected, actual);
    }

    @Test
    public void createInvoiceWithWrongAttributes() {

        try {
            manager.createInvoice(null);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }

        // Invoice have already assigned some id
        Invoice invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        invoice.setId(1l);
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        // Invoice.number is null
        invoice = newInvoice(
                null,"Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // Invoice.customer is null
        invoice = newInvoice(
                "00012/2012",null, date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // InvoiceRow.description is null
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow(null, BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // InvoiceRow.amount is null
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", null, BigDecimal.valueOf(219)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // InvoiceRow.amount is negative number
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", BigDecimal.valueOf(-1, 3), BigDecimal.valueOf(219)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }
        
        // InvoiceRow.unitPrice is null
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), null));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // InvoiceRow.unitPrice is negative number
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(-1,2)));
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (ValidationException ex) {
            //OK
        }

        // InvoiceRow.invoice is null
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        invoice.getRows().get(0).setInvoice(null);
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        // InvoiceRow.invoice points to another invoice
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));        
        Invoice i2 = newInvoice(
                "00034/2012","První stavební", date("2012-08-12"),
                newInvoiceRow("Cement", BigDecimal.valueOf(25), BigDecimal.valueOf(8925,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(1234, 3), BigDecimal.valueOf(219)));
        invoice.getRows().get(0).setInvoice(i2);
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        // InvoiceRow have already assigned some id
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        invoice.getRows().get(0).setId(1000l);
        try {
            manager.createInvoice(invoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        // Check that previous tests didn't affect data in database
        assertTrue(manager.findAllInvoices().isEmpty());

        // these variants should be ok
        invoice = newInvoice(
                "00012/2012","Poctivý & spol.", null,
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        manager.createInvoice(invoice);
        Invoice result = manager.getInvoice(invoice.getId()); 
        assertNotNull(result);
        assertInvoiceDeepEquals(invoice, result);

        invoice = newInvoice(
                "00013/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", BigDecimal.ZERO, BigDecimal.valueOf(219)));
        manager.createInvoice(invoice);
        result = manager.getInvoice(invoice.getId()); 
        assertNotNull(result);
        assertInvoiceDeepEquals(invoice, result);

        invoice = newInvoice(
                "00014/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.ZERO));
        manager.createInvoice(invoice);
        result = manager.getInvoice(invoice.getId()); 
        assertNotNull(result);
        assertInvoiceDeepEquals(invoice, result);

        // Invoice with no rows       
        invoice = newInvoice(
                "00015/2012","Poctivý & spol.", date("2012-08-11"));
        manager.createInvoice(invoice);
        result = manager.getInvoice(invoice.getId()); 
        assertNotNull(result);
        assertInvoiceDeepEquals(invoice, result);

    }

    @Test
    public void updateInvoice() throws SQLException {

        Invoice invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        Invoice i2 = newInvoice(
                "00034/2012","První stavební", date("2012-08-12"),
                newInvoiceRow("Cement", BigDecimal.valueOf(25), BigDecimal.valueOf(8925,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(1234, 3), BigDecimal.valueOf(219)));

        manager.createInvoice(invoice);
        manager.createInvoice(i2);
        Long invoiceId = invoice.getId();
        Invoice result;
        
        invoice = manager.getInvoice(invoiceId);
        invoice.setNumber("00013/2012");
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        invoice.setCustomer("Nástupce poctivého");
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        invoice.setIssued(date("2012-01-02"));
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        InvoiceRow insertedRow = newInvoiceRow("Lepenka", BigDecimal.valueOf(10), BigDecimal.valueOf(12523,2));
        invoice.getRows().add(insertedRow);
        insertedRow.setInvoice(invoice);
        assertNull(insertedRow.getId());
        manager.updateInvoice(invoice);        
        assertNotNull(insertedRow.getId());
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        InvoiceRow deletedRow = invoice.getRows().remove(0);
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);
        assertInvoiceRowIsDeleted(deletedRow);

        invoice = manager.getInvoice(invoiceId);
        invoice.getRows().get(0).setDescription("Něco jiného");
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        invoice.getRows().get(0).setAmount(BigDecimal.valueOf(12));
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        invoice = manager.getInvoice(invoiceId);
        invoice.getRows().get(0).setUnitPrice(BigDecimal.valueOf(1234,2));
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        // Invoice with no rows 
        invoice = manager.getInvoice(invoiceId);
        invoice.getRows().clear();
        manager.updateInvoice(invoice);        
        result = manager.getInvoice(invoiceId);
        assertInvoiceDeepEquals(invoice, result);

        // Check if updates didn't affected other records
        assertInvoiceDeepEquals(i2, manager.getInvoice(i2.getId()));
    }

    @Test
    public void updateInvoiceWithWrongAttributes() {

        Invoice invoice = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));

        manager.createInvoice(invoice);
        Long invoiceId = invoice.getId();
        
        try {
            manager.updateInvoice(null);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }
        
        try {
            invoice = manager.getInvoice(invoiceId);
            invoice.setId(null);
            manager.updateInvoice(invoice);        
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        try {
            invoice = manager.getInvoice(invoiceId);
            invoice.setId(invoiceId - 1);
            manager.updateInvoice(invoice);        
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        try {
            invoice = manager.getInvoice(invoiceId);
            InvoiceRow insertedRow = newInvoiceRow("Lepenka", BigDecimal.valueOf(10), BigDecimal.valueOf(12523,2));
            invoice.getRows().add(insertedRow);
            insertedRow.setInvoice(invoice);
            insertedRow.setId(1000l);
            manager.updateInvoice(invoice);        
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        // We know, that checking of other potencial problems in entity is done
        // in method that is common for creating and updating, therefore is not 
        // necessary to test these cases duplicitly.
    }

    @Test
    public void deleteInvoice() throws SQLException {

        Invoice i1 = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        Invoice i2 = newInvoice(
                "00034/2012","První stavební", date("2012-08-12"),
                newInvoiceRow("Cement", BigDecimal.valueOf(25), BigDecimal.valueOf(8925,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(1234, 3), BigDecimal.valueOf(219)));
        manager.createInvoice(i1);
        manager.createInvoice(i2);
        
        assertNotNull(manager.getInvoice(i1.getId()));
        assertNotNull(manager.getInvoice(i2.getId()));

        manager.deleteInvoice(i1);
        
        assertNull(manager.getInvoice(i1.getId()));
        assertNotNull(manager.getInvoice(i2.getId()));
        
        // Check if also invoice rows are correctly deleted.
        for (InvoiceRow row : i1.getRows()) {
            assertInvoiceRowIsDeleted(row);
        }
    }
    
    private void assertInvoiceRowIsDeleted(InvoiceRow row) throws SQLException {

        Connection connection = null;
        PreparedStatement st = null;
        try {
            connection = ds.getConnection();
            st = connection.prepareStatement(
                "SELECT COUNT(*) FROM INVOICEROW WHERE ID = ?");
            st.setLong(1, row.getId());
            ResultSet rs = st.executeQuery();
            rs.next();
            assertTrue("InvoiceRow with id=" + row.getId() + " is not deleted",rs.getInt(1) == 0);
        } finally {
            DBUtils.closeQuietly(connection, st);
        }
        
    }
    
    @Test
    public void deleteInvoiceWithWrongAttributes() {

        Invoice i1 = newInvoice(
                "00012/2012","Poctivý & spol.", date("2012-08-11"),
                newInvoiceRow("Voda", BigDecimal.valueOf(13234, 3), BigDecimal.valueOf(2452,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(4234, 3), BigDecimal.valueOf(219)));
        Invoice i2 = newInvoice(
                "00034/2012","První stavební", date("2012-08-12"),
                newInvoiceRow("Cement", BigDecimal.valueOf(25), BigDecimal.valueOf(8925,2)),
                newInvoiceRow("Písek", BigDecimal.valueOf(1234, 3), BigDecimal.valueOf(219)));
        
        manager.createInvoice(i1);
        manager.createInvoice(i2);

        List<Invoice> invoices = Arrays.asList(i1,i2);
        
        try {
            manager.deleteInvoice(null);
            fail();
        } catch (IllegalArgumentException ex) {
            //OK
        }

        Invoice testInvoice = manager.getInvoice(i1.getId());
        
        try {
            testInvoice.setId(null);
            manager.deleteInvoice(testInvoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }

        try {
            testInvoice.setId(i2.getId() + 100);
            manager.deleteInvoice(testInvoice);
            fail();
        } catch (IllegalEntityException ex) {
            //OK
        }        
        
        // Check that previous tests didn't affect data in database
        assertInvoiceCollectionDeepEquals(invoices, manager.findAllInvoices());
    }
    
    private static Invoice newInvoice(String number, String customer, Date issued, InvoiceRow ... rows) {
        Invoice invoice = new Invoice();
        invoice.setNumber(number);
        invoice.setCustomer(customer);
        invoice.setIssued(issued);
        for (InvoiceRow row : rows) {
            invoice.getRows().add(row);
            row.setInvoice(invoice);
        }
        return invoice;
    }

    private static InvoiceRow newInvoiceRow(String description, BigDecimal amount, BigDecimal unitPrice) {
        InvoiceRow invoiceRow = new InvoiceRow();
        invoiceRow.setDescription(description);
        invoiceRow.setAmount(amount);
        invoiceRow.setUnitPrice(unitPrice);
        return invoiceRow;
    }    
   
    
    private void assertInvoiceDeepEquals(Invoice expected, Invoice actual) {
        assertEquals(expected.getId(), actual.getId());
        assertEquals(expected.getNumber(), actual.getNumber());
        assertEquals(expected.getCustomer(), actual.getCustomer());
        assertEquals(expected.getIssued(), actual.getIssued());
        assertEquals(expected.getRows(), actual.getRows());
        for (int i = 0; i < expected.getRows().size(); i++) {
            InvoiceRow expectedRow = expected.getRows().get(i);
            InvoiceRow actualRow = actual.getRows().get(i);
            assertEquals(expectedRow.getId(), actualRow.getId());
            assertEquals(expectedRow.getInvoice(), actualRow.getInvoice());
            assertEquals(expectedRow.getDescription(), actualRow.getDescription());
            assertEquals(expectedRow.getAmount(), actualRow.getAmount());
            assertEquals(expectedRow.getUnitPrice(), actualRow.getUnitPrice());
        }
    }

    private Comparator<Invoice> invoiceKeyComparator = new Comparator<Invoice>() {

        @Override
        public int compare(Invoice o1, Invoice o2) {
            Long k1 = o1.getId();
            Long k2 = o2.getId();
            if (k1 == null && k2 == null) {
                return 0;
            } else if (k1 == null && k2 != null) {
                return -1;
            } else if (k1 != null && k2 == null) {
                return 1;
            } else {
                return k1.compareTo(k2);
            }
        }
    };
    
    private void assertInvoiceCollectionDeepEquals(List<Invoice> expected, List<Invoice> actual) {
        assertEquals(expected.size(), actual.size());
        List<Invoice> expectedSortedList = new ArrayList<Invoice>(expected);
        List<Invoice> actualSortedList = new ArrayList<Invoice>(actual);
        Collections.sort(expectedSortedList,invoiceKeyComparator);
        Collections.sort(actualSortedList,invoiceKeyComparator);
        for (int i = 0; i < expectedSortedList.size(); i++) {
            assertInvoiceDeepEquals(expectedSortedList.get(i), actualSortedList.get(i));            
        }
    }
}
